#! /usr/bin/python  -O

# Packet Garden master script (beta). Cross platform version 
# 
# Copyright (C) 2006 Julian Oliver
#
# This script is responsible for the user interface part of Packet Garden
# and manages most of the project's functions. The widget library used, 
# soya.Pudding, is 'quirky' to say the least. For this reason you'll find that
# approaches to menu setup look pretty inconsistent in parts: some menus 
# are created in a given class's __init__ while others aren't, some
# menus have their own container while others are added to the root widget.
# That's love on the Freeway.
#
# License:
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import os
import soya
import soya.laser
import soya.pudding as pudding
import PIL.Image as pil
import soya.pudding.ext.slicingimage
import pg_player 
import soya.pudding.ext.meter
import time
import random
import math
import socket
import subprocess
import urllib
import re
import sets
import tempfile
import gzip
import textwrap
# import soya.particle

try:
  import GeoIP
except ImportError:
  print "GeoIP module not found. there will be no country codes generated for captured IP's"
  pass
try:
  import struct
except ImportError:
  pass

try:
  import psyco
  psyco.full()
except ImportError:
  print "psyco module not found. you're probably on a PowerPC. performance will be reduced. sorry!"
  pass

PLATFORM = sys.platform

if "darwin" in PLATFORM:
  PG_DIR = os.curdir 
  from Authorization import Authorization, kAuthorizationFlagDestroyRights
elif "win32" in PLATFORM:
  import PIL.PngImagePlugin
  import PIL.JpegImagePlugin
  import win32com.client
  PG_DIR = os.curdir 
else:
  PG_DIR = os.path.join(os.path.expanduser('~/'), '.packetgarden/')
soya.path.append(os.path.join(os.path.dirname(PG_DIR), "data"))

hostname = socket.gethostname()

# check user config for fullscreen option. if this file doesn't exist, create it.
try:
  user_cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'r')
  options = user_cfg.read()
  if re.search("fullscreen on", options):
    soya.init("Packet Garden", 1024, 768, 1)
  else:
    soya.init("Packet Garden", 1024, 768, 0, 0) # some seem to fullscreen w/o explicit toggle.
except:
  user_cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'w')
  user_cfg.write("fullscreen off" + '\n'+"shadows off" + '\n' +"generation filter:0.1")
  user_cfg.close()
  soya.init("Packet Garden", 1024, 768, 0, 0)

# close config file for re-use later
user_cfg.close()

date = time.strftime('%d'+'-'+'%m'+'-'+'%y')

pudding.init()
scene = soya.World()
camera = soya.Camera(scene)
camera.set_xyz(0.0, 100.0, 0.0)
center = soya.Point(scene, 0.0, 0.0, 0.0)

p = pg_player.avatar
url = urllib.URLopener()

class text_menu(pudding.control.Label):
  def __init__(self, label, function, font, arg):
    pudding.control.Label.__init__(self, label = label, font = font)
    self.arg = arg
    self.function = function
    self.color1 = (0.8, 0.8, 0.8, 0.5)
    self.color = self.color1
    self.color2 = (0, 1, 0, 1)
  def on_focus(self):
    self.color = self.color2
  def on_loose_focus(self):
    self.color = self.color1
  def on_mouse_up(self, button, x, y):
    self.function(self.arg)
    return True # check if this is necessary. 
  
class button_menu(pudding.control.Button):
  def __init__(self, width, height, top, left, function, label, arg):
    pudding.control.Button.__init__(self, height = height, width = width, top = top, left = left, label = label)
    self.arg = arg
    self.label = label
    self.function = function
    self.border_color = (1, 1, 1, .2)
    self.background_color = (0.1, 0.1, 0.1, 0.0)
  def on_focus(self):
    self.border_color = (0.1, 0.5, 0.1, 1)
  def on_loose_focus(self):
    self.border_color = (1, 1, 1, .2)
  def on_mouse_up(self, button, x, y):
    self.function(self.arg)
    return True # check if is this necessary.

class slider(pudding.control.Box):
  ''' 
  a value slider. 
  '''
  def __init__(self, label, init_val, range):
    pudding.control.Box.__init__(self, height = 22, width = 32, top = 344, left = 0)
    self.label = label
    self.init_val = init_val
    self.pad = 4
    self.left = ((self.init_val)*range)+(range/2)+self.pad
    self.left_limit = (self.width*1.5)
    self.right_limit = (self.width*1.5) + range +1 # tweak tweak
    self.label.left = self.left + 6
    self.range = (self.right_limit - self.left_limit)
    self.val = self.init_val
    self.val_min = 0.01 # 0.001 is our sane minumum.
    self.val_max = 0.99 # 1.0 would mean no generation 
  def on_focus(self):
    self.border_color = (0.1, 0.5, 0.1, 1)
  def on_loose_focus(self):
    self.border_color = (1, 1, 1, .2)
    soya.set_grab_input(0)
  def on_mouse_down(self, button, x, y):
    soya.set_grab_input(1)
  def on_mouse_up(self, button, x, y):
    soya.set_grab_input(0)
  def advance_time(self, proportion):
    mouse_pos_x = soya.get_mouse_rel_pos()[0]
    if self.left >= self.left_limit and self.left <= self.right_limit and soya.get_grab_input() == 1:
      self.left += mouse_pos_x # move slider puck to relative mouse pos 
      self.label.left = self.left +self.width/4
      relative_pos = self.range - (self.right_limit - self.left) # find relative position of puck inside limits
      self.val = round(relative_pos/(self.range), 3) # generate value from 0 - 1, round to three decimal places
    elif self.val <= self.val_min: 
      soya.set_grab_input(0)
      self.left = self.left_limit+1
      self.val = 0.01 # safe lower limit
    elif self.val >= self.val_max: 
      soya.set_grab_input(0)
      self.left = self.right_limit-1
      self.val = .99
    elif soya.get_grab_input() == 0:
      self.left += 0 
    else:
      self.left += 0
    self.label.label = str(self.val*100)[:-2] # update label value as a percentage each cycle.

class configurator:
  '''
  configuration interface class. this is setup in a lavishly dumb way. must fix.
  '''
  def __init__(self):
    user_cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'r')
    options = user_cfg.readlines()
    if re.search("shadows off", options[0]):
      self.shadow_state = "shadows off"
    else:
      self.shadow_state = "shadows on"
    if re.search("fullscreen off", options[1]):
      self.window_state = "fullscreen off"
    else:
      self.window_state = "fullscreen on"
    self.generation_filter = "generation filter:"
    a, b = options[2].split(':')
    self.gen_filter = float(b.strip("\n"))
    self.shadow_msg = "turn on shadows if you have a powerful graphics card."
    self.window_msg = "turn on/off fullscreen mode."
    self.gen_msg = "drag slider to change percentage of captured data to be filtered out (0.1% - 99%).\n\
lower values let through more detail from captured traffic but increase generation time."
    self.save_msg = "changes will be applied after Packet Garden restarts."
    self.shadow_toggle = 0
    self.window_toggle = 0
    # keep this out of setup() so that the value can persist
    self.slider_label = pudding.control.SimpleLabel(widget, label = "50", font=font1, top=346, left=100, z_index = 5)
    self.slider_puck = widget.add_child(slider(self.slider_label, self.gen_filter, 88))
    self.slider_puck.visible = 0
    self.slider_label.visible = 0
    self.setup()
    self.cont.visible = 0
  def setup(self): # sets up everything in menu. most reliable 'redraw' i could get without everything exploding.
    self.cont = pudding.container.VerticalContainer(widget, width = 640, height = 480, left = 10, top = 210)
    self.shadow_label = pudding.control.SimpleLabel(self.cont, label = self.shadow_state, font=font1, top=35, left=60)
    self.shadow_about = pudding.control.SimpleLabel(self.cont, label = self.shadow_msg, font=font1, top=35, left=180, color = (1, 1, 1, .3))
    self.window_label = pudding.control.SimpleLabel(self.cont, label = self.window_state, font=font1, top=85, left=60)
    self.window_about = pudding.control.SimpleLabel(self.cont, label = self.window_msg, font=font1, top=85, left=180, color = (1, 1, 1, .3))
    self.gen_label = pudding.control.SimpleLabel(self.cont, label = self.gen_msg, font=font1, top=130, left=180, color = (1, 1, 1, .3))
    self.save_label = pudding.control.SimpleLabel(self.cont, label = "save and quit", font=font1, top=185, left=60)
    self.save_about = pudding.control.SimpleLabel(self.cont, label = self.save_msg, font=font1, top=185, left=180, color = (1, 1, 1, .3))
    self.box = pudding.control.Box(self.cont, width = 960, height = 200, top = 20, left = 20, \
        background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .1))
    self.slider_box = pudding.control.Box(self.cont, width = 128, height = 30, top = 130, left = 35, \
        background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .2))
    self.cont.visible = 1
  def menu(self, arg):
    self.setup()
    self.cont.add_child(button_menu(128, 30, 30, 35, self.shadows_button, "", None))
    self.cont.add_child(button_menu(128, 30, 80, 35, self.window_button, "", None))
    self.cont.add_child(button_menu(128, 30, 180, 35, self.save, "", None))
    visibility([menu_cont, menu_box, selector.up_cont, selector.down_cont, selector.arrow_box, selector.cont, selector.box, selector.text], 0)
    visibility([back_label, back_cont, self.slider_puck, self.slider_label], 1)
  def shadows_button(self, arg):
    self.shadow_toggle +=1
    if self.shadow_toggle == 1:
      self.shadow_state = "shadows off"
      self.shadow_label.label = self.shadow_state
    else:
      self.shadow_toggle = 0
      self.shadow_state = "shadows on"
      self.shadow_label.label = self.shadow_state
  def window_button(self, arg):
    self.window_toggle +=1
    if self.window_toggle == 1:
      self.window_state = "fullscreen off"
      self.window_label.label = self.window_state
    else:
      self.window_toggle = 0
      self.window_state = "fullscreen on"
      self.window_label.label = self.window_state
  def save(self, arg):
    self.cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'w')
    self.cfg.write(self.shadow_state +'\n')
    self.cfg.write(self.window_state +'\n')
    self.cfg.write(self.generation_filter +str(self.slider_puck.val) +'\n')
    self.cfg.close() 
    time.sleep(1)
    soya.quit()
    if "darwin" in PLATFORM:
	  os.execl('/bin/ls','/bin/ls')
	  #os.execl('/bin/sh', '/bin/sh',os.path.join(pg_dir, 'packetgarden'))
      #os.execl(os.path.join(os.path.dirname(__file__),'packetgarden'))
    elif "win32" in PLATFORM:
      os.execl('pg_garden.exe')
    else:
      try:
        os.execl('/usr/share/games/packetgarden/pg_garden.py')
      except:
        os.execl('python ./pg_garden.py')

class make_world:
  '''
  master class to make our world. 
  '''
  def __init__(self):
    self.eta = 0
    self.peak = []
    self.count = 0
    self.garden = soya.World()
    self.preview = soya.World()
    self.label_world = soya.World()
    self.gen_plant_world = soya.World()
    self.point = soya.Point(soya.World(), 0.0,0.0,0.0)
    self.cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'r').readlines()
    f = self.cfg[2].split(':')[1].strip()
    self.filter = float(f)
    if "shadows off\n" in self.cfg:
      self.suffix = '_no-shadow'
      self.shadow_state = 0
    else:
      self.suffix = ""
      self.shadow_state = 1
    self.setup()
  def setup(self):
    self.prev_label = pudding.control.SimpleLabel(widget, label = "no plants were generated for this world", \
      autosize = True, font=font1, top=40, left=30, color= (1, 1, 1, .3))
    self.prev_label.visible = 0
    self.cont = pudding.container.VerticalContainer(widget, width = 10, height = 50, left = 30, top = 240, z_index=-4)
    self.date_label  = pudding.control.SimpleLabel(widget, label = "", font=font1, top=25, left=30, color = (1, 1, 1, .6))
    self.box = pudding.control.Box(self.cont, width = 960, height = 220, top = 10, left = 0, \
      background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .1))
    self.count_label = pudding.control.SimpleLabel(self.cont, label = "", font=font1, top=25, left=20, z_index=2)
    self.meter = pudding.ext.meter.Meter(self.cont, min = 0, max = 100, left=20, top=45, width=520, height=20, z_index=-1)
    self.stats_label = pudding.control.SimpleLabel(self.cont, label = "", font=font1, top=65, left=20)
    #self.meter.color = (.4, .4, .4, 1) # white
    self.meter.color = (.1, .5, .1, 1) # green
    self.meter.border_color = (.3, .3, .3, .1)
    self.stats_label.color = (1, 1, 1, .4)
    self.count_label.color = (1, 1, 1, .4)
  def run(self, filter):
    '''
    here's our main terrain creation code. it reads in entries from the log, orders them in a list and
    plots them using a spherical coord system derived from the first two blocks of the IP itself.

    notes on value representation in the terrain graph:

    if you download a 750Mb video and then download a 100k image from Flickr.com you'd need a massive 
    scale before the 100k download of image data would be properly comparitively portrayed in the 
    graph at 0.00013r the size, perhaps equivalent to a park bench before a skyscraper.

    for this reason i've had to play around with alternative representations to avoid single-peak worlds.

    the first i explored was the Sigmoid function (on pix's <http://pix.test.at> suggestion), giving emphasis 
    to lower values.

    it worked out ok, but it was difficult to read the graph when on the world surface, and often produced worlds 
    which felt 'distorted' somehow.

    for this reason decided to go for a linear distribution model based on list ordering as 
    - while also not being value-representative - it does allow very small values to have visible presence
    alongside high values. in the graph, peaks are then read as "this peak is smaller than that" as opposed to 
    "this peak is smaller than that by this proportion".
    '''
    self.setup()
    self.external_ip, self.ip_check = get_external_ip()
    self.today = time.strftime('%d'+'-'+'%m'+'-'+'%y')
    try:
      if os.path.join(PG_DIR, 'labels', 'pg_labels_' +self.today +'.labels'):
        os.remove(os.path.join(PG_DIR, 'labels', 'pg_labels_' +self.today +'.labels'))
    except:
      pass
    self.labels_file = open(os.path.join(PG_DIR, 'labels/', 'pg_labels_' +self.today +'.labels'), 'a')
    self.labels = {}
    self.cont.visible = 1
    self.t = time.time()
    geom = open(os.path.join(PG_DIR, 'data', 'sphere-big.txt'), 'r')
    #geom = open(os.path.join(PG_DIR, 'data', 'sphere.txt'), 'r')
    self.world = soya.World()
    for lines in geom:
      lines = lines.strip()
      if len(lines):
        x,y,z = map(float, lines.split(',')) 
        f.append(soya.Vertex(self.world, x, y, z, diffuse = (.2, .2, 0.0, 1)))
      else:
        f = soya.Face(self.world)
        f.double_sided = 1
    menu_toggle(0)
    self.box.visible = 1
    self.filter = filter
    try:
      # if it can't find this file it excepts which raises a message
      pcap_log  = gzip.open(os.path.join(PG_DIR, 'logs', 'pcap_' +hostname +'_' +str(time.strftime('%d'+'-'+'%m'+'-'+'%y'))+'.log.gz'), 'r')
      #pcap_log  = gzip.open(os.path.join(PG_DIR, 'logs', 'demo.log.gz'), 'r')
      for lines in pcap_log:
        lines = lines.strip()
        lines = lines.strip("(").strip(")")
        ip, dir, type, region, length = str.split(lines, ", ")
        x, y, z, w = ip.split('.')
        block1, block2, region, ip, length = int(x.strip("'")), int(y.strip("'")), region.strip("'"), \
        ip.strip("'"), int(length)
        self.peak.append([ip, block1, block2, length, dir, type, region])
      # add our external IP to the list
      limit_list = [float(i[3]) for i in self.peak]
      self.maxval = max(limit_list)
      self.height = int(.01 * self.maxval) # give ourselves a height
      if self.ip_check == 1:
        x, y, z, w = self.external_ip.split('.')
        block1, block2 = int(x), int(y)
        self.peak.append([self.external_ip.strip(), block1, block2, self.height, 'up', 'None', 'home'])
      key = lambda i: i[3]
      self.peak.sort(key=key) # thanks Juhaz, MFen @ #python for this lovely in-place sort by key;) 
      self.max_megs = str((self.filter*self.maxval)/1048576.0)
      self.labels_file.write(str(self.maxval) +'\n')
      radius = 1.401
      vecs = {}
      proc_total = []
      boost = 50
      self.peak_len = len(self.peak)
      for i in self.peak:
          if i[6] == 'home' or (self.peak.index(i)/float(self.peak_len)) > self.filter:
          #if i[6] == 'home' or i[3] > self.filter * self.maxval:
            proc_total.append(i)
      self.proc_total = len(proc_total)
      for i in self.peak:
        #if i[3]: # every logged IP. slow.
        if i[6] == 'home' or (self.peak.index(i)/float(self.peak_len)) > self.filter:
          # progress bar output
          self.eta = 100 - (((self.proc_total - self.count)/float(self.proc_total))*100) # range of progress bar in pg_garden.py
          self.meter.__set_value__(self.eta)
          self.count_label.label = "Processing " +str(self.count) +" of " +str(self.proc_total) +" transactions greater than " +self.max_megs +" megabytes"
          if PLATFORM == "win32" and i[6] != 'home':
            self.stats_label.label ="Address: "      +str(i[0] +'\n')\
                                    +"Protocol: "      +str(i[5] +'\n')\
                                    +"Direction: "      +str(i[4] +'\n')\
                                    +"Transfer amount: " +str(i[3]/1048576.0) +" megabytes" # this data feeds the progress bar
          else:
            if i[6] != 'home':
              self.stats_label.label ="Country: "+str(i[6] +'\n')\
                                      +"Address: "      +str(i[0] +'\n')\
                                      +"Protocol: "      +str(i[5] +'\n')\
                                      +"Direction: "      +str(i[4] +'\n')\
                                      +"Transfer amount: " +str(i[3]/1048576.0) +" megabytes" # this data feeds the progress bar
            else:
              pass
          pudding.main_loop.MainLoop(scene).update()
          self.count += 1
          angX, angZ = float(i[1]/256.0), float(i[2]/256.0)
          #height = boost/(boost+(float(i[3])/self.maxval)**(-1.0)) # Sigmoid Function with boost
          base_height =   self.peak.index(i)/float(self.peak_len) # value created from order in list.
          height = (base_height-self.filter)/(1-self.filter) # re-scale between 0 and 1
          if i[4] == 'up': # switch direction for uploads
            height = height * -1.0 
          angX = (2*math.pi) * angX
          angZ = math.pi * angZ
          # spherical coordinate translations
          self.point.set_xyz(radius* (math.cos(angX) * math.sin(angZ)), radius* (math.sin(angX) * math.sin(angZ)), radius* (math.cos(angZ)))
          vec = self.point.position().vector_to(self.world.position()) # vector of self.peak
          for face in self.world: 
            for vert in face:
              v = vert.position()
              w = self.world.position()
              dw = v.distance_to(w)
              dv = v.vector_to(w).distance_to(vec)
              p = v.distance_to(self.point.position()) 
              if dv < .25 and p < radius: 
                dist = 0.25 - dv 
                vert.add_mul_vector(dist*height, vec) # smoothing
                vecs[dv] = vert                 
              else:
                vecs[dv] = vert
              v = vert.position()# have to call this again once the vert's moved before colouring it
              dw = v.distance_to(w) 
              if dw > 1.6:
                vert.color = (0.6, 0.6, 1.0, 1.0) 
              elif dw < 1.3:
                vert.color = (0.3, 0.2, 0.0, 1.0) #brown
              else:
                pass
          try:
            center = vecs[min(vecs.keys())]
            # we need this to collate in a dict first. cannot write to file directly.
            self.labels[i[0] +", " +i[5] +", " +i[4] +", " +str(i[3]) +", " +i[6]] = center 
            vecs.clear()
          except:
            print "broken labels"
            pass
      hour, minute =time.localtime()[3], time.localtime()[4]
      if hour < 10:
        hour = "0"+str(time.localtime()[3])
      else:
        hour = str(time.localtime()[3])
      if minute < 10:
        minute = "0"+str(time.localtime()[4])
      else:
        minute = str(time.localtime()[4])
      for k, v in self.labels.iteritems():
        self.labels_file.write(str(k) +" | " +str(v.x) +", " +str(v.y) +", " +str(v.z) +'\n')
      self.labels_file.close()
      self.world.set_scale_factors(30, 30, 30)
      self.world.filename = 'packet-garden_' +str(time.strftime('%d'+'-'+'%m'+'-'+'%y'))
      self.world.save()
      self.make_garden()
    except:
      self.stats_label.label = "oops, nothing to process!\n\
please start packet capture from the main menu \n\
and generate some data by using the internet."
      self.meter.visible = 0
      back_cont.visible = 1
      back_label.visible = 1
      pudding.main_loop.MainLoop(scene).update()
      #menu_toggle(1)
  def make_stats(self):
    visibility([self.meter, self.count_label], 0)
    data_list = []
    zone_list = []
    country_list = []
    up_total = 0
    down_total = 0
    mail_count = secure_count = p2p_count = im_count = other_count = http_count = game_count = ftp_count = 0
    root_zones = open(os.path.join(PG_DIR, 'stats', 'root-zones.txt'), 'r')
    for lines in root_zones:
      zone_list.append([lines.split(" | ")])
    self.now = str(time.strftime('%d'+'-'+'%m'+'-'+'%y'))
    self.labels_list = labels_file = open(os.path.join(PG_DIR, 'labels', 'pg_labels_' +self.now +'.labels'), 'r').readlines()
    self.maxval = self.labels_list[0]
    self.maxval = self.maxval.rstrip('\n')
    self.maxval = float(self.maxval)
    for i in self.labels_list[1:]:
      label, pos = str.split(i, " | ")
      ip, port, dir, length, region = str.split(label, ", ")
      port.strip(","); port = port.strip()
      dir.strip("'"); dir = dir.strip() #do we need this?
      length.strip("'"); length.strip()#do we need this?
      data_list.append(port); data_list.append(dir), data_list.append(region.swapcase())
      if dir == 'up':
        up_total += int(length)
      if dir == 'down':
        down_total += int(length)
    for i in ['smtp', 'imap', 'pop3']:
      mail_count += data_list.count(i)
    for i in ['ssh', 'https']:
      secure_count += data_list.count(i)
    for i in ['mule', 'gnutella', 'soulseek']:
      p2p_count += data_list.count(i)
    for i in ['icq', 'msn', 'bt']:
      im_count += data_list.count(i)
    for i in ['quake3', 'ut2004']:
      game_count += data_list.count(i)
    other_count = data_list.count('other')
    http_count = data_list.count('http')
    ftp_count = data_list.count('ftp')
    up_count = data_list.count('up')
    down_count = data_list.count('down')
    countries = c = ""
    txt_wrap = textwrap.TextWrapper(width=44)
    for i in zone_list:
      if i[0][0] in data_list:
        c = c +i[0][0].swapcase()+" "
    if not "win32" in PLATFORM:
      l = txt_wrap.wrap(c)
      for i in l:
        countries = countries +i +'\n'+" "*17
      country_string = "Countries visited: "
    else:
      country_string = ""
    stats_msg = "Ignored all traffic less than " +self.max_megs +" megabytes." '\n' \
                +"Uploaded: " +str(up_total/1048576.0) +" megabytes " +"to " +str(up_count) +" computers" '\n' \
                +"Downloaded: "+str(down_total/1048576.0) +" megabyes " +"from " +str(down_count) +" computers" '\n' '\n' \
                +"Securely exchanged data with " +str(secure_count) +" computers"'\n' \
                +"Sent or received mail via " +str(mail_count) +" computers"'\n' \
                +"Shared files with " +str(p2p_count) +" computers"'\n' \
                +"Transferred files using FTP with " +str(ftp_count) +" computers" '\n' \
                +"Played games on " +str(game_count) +" servers" '\n' \
                +"Sent or received instant messages via " +str(im_count) +" computers" '\n' \
                +"Unknown traffic was exchanged with " +str(other_count) +" computers" '\n' \
                +country_string +countries
    data = open(os.path.join(PG_DIR, 'stats', 'pg_stats_' +self.now +'.sheet'), 'w')
    data.write(stats_msg)
    data.close()
    menu_toggle(0)
    self.stats_label.label = stats_msg
    self.stats_label.top = 20
    self.stats_label.color = (1, 1, 1, .4)
    pudding.main_loop.MainLoop(scene).update()
  def generate(self, args):
    self.run(self.filter)
  def make_garden(self):
    self.make_stats()
    self.add_plants(self.gen_plant_world, self.labels_list[1:], self.maxval, self.suffix)
    self.gen_plant_world.filename = 'plants_' +self.now
    self.gen_plant_world.save()
    self.previewer((self.now, 0))
  def previewer(self, (date, toggle)):
    print self.shadow_state
    # make sure preview worlds aren't added twice
    self.toggle = toggle 
    try:
      for i in scene.recursive():
        if type(i) == soya.Light:
          scene.remove(i)
    except:
      pass
    try: 
      self.prev_label.visible = 0
      self.preview.advance_time = lambda r: self.preview.rotate_y(0*r) # reset rotation to 0 speed 
      for i in self.preview.recursive():
        self.preview.remove(i)
      scene.remove(self.preview)
      try: # remove plants if they were generated. must be removed separately.
        self.preview.remove(self.plant_world)
      except:
        print "removing plants in previewer failed"
        pass
    except:
      pass
    if os.path.exists(os.path.join(PG_DIR, 'data', 'worlds', 'packet-garden_' +date + '.data')):
      self.world_shape = soya.World.get('packet-garden_' +date).shapify()
      self.world = soya.Body(self.preview, self.world_shape)
    else:
      print "more than just 'didn't work'" # some useful error output for debugging purposes.
    if os.path.exists(os.path.join(PG_DIR, 'data', 'worlds', 'plants_' +date +'.data')):
      self.plant_world = soya.World.get('plants_' +date)
      self.preview.add(self.plant_world)
    else:
      self.prev_label.visible = 1
      print "No plants were generated for this world."
    if date == 'demo':
      self.date_label.label = "demo garden"
    else:
      self.date_label.label = "garden created on: " +date 
    if self.shadow_state == 1:
      self.reserved_world = soya.World.get('reserved_world')
    else:
      self.reserved_world = soya.World.get('reserved_world_no-shadow')
    self.preview.add(self.reserved_world)
    camera.set_xyz(0.0, 0.0, 100.0)
    sun = soya.Light(scene) # consider removing this sun before visiting world?
    sun.directional = 0
    sun.diffuse = (1, 1, 1, 1)
    sun.set_xyz(0, 300, 300)
    #sun.shadow_color = (0.0, 0.0, 0.1, .4)
    sun.shadow_color = (0.0, 0.0, 0.1, .2)
    atmosphere = soya.SkyAtmosphere()
    atmosphere.ambient = (0.3, 0.3, 0.3, 1)
    atmosphere.skyplane  = 1
    atmosphere.sky_color = (1, 1, 0.8, 1)
    scene.atmosphere = atmosphere
    self.preview.advance_time = lambda r: self.preview.rotate_y(.3*r) # shamelessly ripped off from jiba
    scene.add(self.preview)
    if self.toggle == 0: # we want generated worlds to be displayed in all their glory the first time.
      # turn off menus and buttons before creating our thumbnail.
      visibility([back_cont, back_label, quit_label, quit_cont, selector.cont, selector.box, selector.arrow_box, \
      selector.arrow_down, selector.arrow_up, play_cont, play_label, self.date_label, self.cont, self.stats_label, self.box], 0) 
      menu_toggle(0)
      soya.render()
      soya.screenshot().resize((320, 240)).save(os.path.join(PG_DIR, 'data', 'images', 'thumbnails', 'pg_' \
        +date +'.jpeg'))
      # turn buttons back on after creating thumbnail
      visibility([play_cont, play_label, back_cont, back_label, quit_label, quit_cont, self.date_label, self.cont], 1) 
    else:
      visibility([play_label, play_cont], 1)
    play.arg = date
  def go_world(self, date): 
    menu_toggle(0) # turn off the menus before entering the garden 
    # remove the old rotated reserved world
    try:
      self.preview.remove(self.reserved_world)
    except:
      pass
    visibility([back_cont, back_label, quit_label, quit_cont, selector.up_cont, selector.down_cont, selector.arrow_box, \
      selector.cont, selector.box, selector.text, play_cont, play_label, configure.cont, self.cont], 0) 

    if self.shadow_state == 1:
      self.reserved_world = soya.World.get('reserved_world')
    else:
      self.reserved_world = soya.World.get('reserved_world_no-shadow')
    self.garden.add(self.reserved_world) 
    visibility([game_box, game_label], 1)
    if date == None:
      date = time.strftime('%d'+'-'+'%m'+'-'+'%y')
      visibility([back_cont, back_label, quit_label, quit_cont, selector.up_cont, selector.down_cont, selector.arrow_box, \
      selector.cont, selector.box, selector.text, menu_cont, menu_box, self.cont], 0)
    try:
      scene.remove(self.preview)
      self.preview.remove(self.plant_world)
    except:
      pass
    if os.path.exists(os.path.join(PG_DIR, 'data', 'worlds', 'packet-garden_' +date +'.data')):
      self.world_shape = soya.World.get('packet-garden_' +date).shapify()
      self.world = soya.Body(self.garden, self.world_shape)
      self.garden.add(self.world)
    else:
      print "No garden generated for this day!"
    if os.path.exists(os.path.join(PG_DIR, 'data', 'worlds', 'plants_' +date +'.data')):
      self.plant_world = soya.World.get('plants_' +date)
      self.garden.add(self.plant_world)
    else:
      print "No plants were generated."
    for i in  selector.cont.children: i.visible = 0
    scene.add(self.garden)
    sun2 = soya.Light(self.garden)
    sun2.directional = 1
    sun2.diffuse = (1, 1, 1, 0.5)
    sun2.shadow_color = (0.0, 0.0, 0.1, .4)
    player = p(scene, self.garden, self.world_shape, self.label_world, self.plant_world)
    third_person = soya.ThirdPersonTraveling(player)
    third_person.distance = 10.0
    camera = soya.TravelingCamera(scene)
    camera.add_traveling(third_person)
    camera.zap()
    camera.fov = 90
    camera.back = 100.00
    self.add_labels(self.garden, camera, player, date) # write to txt here
    widget.add_child(camera)
    pudding.main_loop.MainLoop(scene).stop()
    time.sleep(.5)
    soya.Idler(scene).idle()
  def add_labels(self, scene, camera, player, date):
    self.scene = scene
    # add reserved block labels and objects at the same time.
    reserved_labels = open(os.path.join(PG_DIR, 'labels', 'reserved.labels'), 'r')
    #reserved_mdl = soya.Shape.get('reserved')
    #reserved_world= soya.World()
    for lines in reserved_labels:
      label, pos = str.split(lines, " | ")
      ip, port, dir, length = str.split(label, ", ")
      x, y, z = str.split(pos, ", ")
      a, b, c, d = str.split(ip, ".")
      self.iplabel = pg_label(self.label_world, float(x), float(y), float(z), str(ip) + " private" , camera, player, 1)
      self.iplabel.label3d.add_mul_vector(-6.2, self.iplabel.vector_to(self.world.position()))
      self.iplabel.label3d.lit = 0
      # self.iplabel.label3d.color=(.5, 1, 0, .3) 
      self.iplabel.label3d.color=(.9, 1.0, 0, .6) # red 
      # self.iplabel.label3d.color=(1, .5, 0, .5) # orange 
      # self.iplabel.label3d.color=(.8, 0, 0, .7) # red
      self.iplabel.label3d.scale(1, 1, 1)
      #reserved_body = soya.Body(reserved_world, reserved_mdl)
      #reserved_body.set_xyz(float(x)*30, float(y)*30, float(z)*30)
      #reserved_body.add_mul_vector(-.25, reserved_body.position().vector_to(self.label_world.position()))
      #reserved_body.look_at(reserved_body.position().vector_to(self.label_world.position()))
    #self.label_world.add(reserved_world)
    #reserved_world.filename = "reserved_world"
    #reserved_world.save()
    if os.path.exists(os.path.join(PG_DIR, 'labels', 'pg_labels_' +str(date) + '.labels')):
      labels = open(os.path.join(PG_DIR, 'labels', 'pg_labels_' +str(date) + '.labels'), 'r')
      labels_list = labels.readlines()[1:]
      for labels in labels_list:
        label, pos = str.split(labels, " | ")
        ip, port, dir, length, region = str.split(label, ", ")
        if port == 'other':
          port = 'unknown'
        x, y, z = str.split(pos, ", ")
        a, b, c, d = str.split(ip, ".")
        self.offset_y = -4.0 +(int(c)/256.0)
        if "win32" in PLATFORM:
          self.iplabel = pg_label(self.label_world, float(x), float(y), float(z), str(a +"." +b +".0.0"), camera, player, 1)
        else:
        # self.offset_x = 1 -(int(d)/256.0) # offset for labels in the same location
          self.iplabel = pg_label(self.label_world, float(x), float(y), float(z), str(a +"." +b +".0.0 (" +region +") "+dir), camera, player, 1)
        self.iplabel2 = pg_label(self.label_world, float(x), float(y), float(z), ip +" " +port, camera, player, 0)
        self.label_parms() 

    self.scene.add(self.label_world)
  def label_parms(self):
      self.iplabel.label3d.add_mul_vector(-4.5, self.iplabel.vector_to(self.world.position()))
      self.iplabel.label3d.scale(1, 1, 1)
      self.iplabel.label3d.color=(1, 1, 1, .2) 
      self.iplabel.label3d.lit = 0
      self.iplabel2.label3d.add_mul_vector(self.offset_y, self.iplabel2.vector_to(self.world.position()))
      self.iplabel2.label3d.color=(1, 1, 1, .2) 
      self.iplabel2.label3d.scale(1, 1, 1)
      self.iplabel2.label3d.lit = 0
  def add_plants(self, world, labels, maxval, suffix):
    self.labels = labels
    suffix = suffix
    self.world = world
    for labels in self.labels:
      self.scene = scene
      label, pos = str.split(labels, " | ")
      ip, port, dir, length, region = str.split(label, ", ")
      x, y, z = str.split(pos, ", ")
      a, b, c, d = str.split(ip, ".")
      port.strip(","); port = port.strip()
      dir.strip("'"); dir = dir.strip()#do we need this?
      length.strip("'"); length.strip()#do we need this?
      scale = 1/(1+self.maxval**((float(length)/self.maxval) * -1)) # thnx Sigmoid.
      make_plant(self.world, float(x), float(y), float(z), str(port), scale, suffix) # create plants 

class world_selector:
  ''' 
  interface to browse and run previously generated worlds
  ''' 
  def __init__(self):
    atmosphere = soya.SkyAtmosphere()
    atmosphere.ambient = (0.3, 0.3, 0.3, 1)
    atmosphere.skyplane  = 1
    atmosphere.sky_color = (1, 1, 0.8, 1)
    scene.atmosphere = atmosphere
    self.setup()
  def get_pics(self, arg, top, names):
    self.pics.append(names)
  def setup(self):
    self.pics = []
    # clean this up. make self.cont a scroll_cont or similar and put all the other bits into a selector_cont
    self.cont = pudding.container.VerticalContainer(widget, width = 800, height = 400) # this container is moved up and down
    self.up_cont = pudding.container.VerticalContainer(widget, width = 10, height = 400, top = 260, left = 920)
    self.down_cont = pudding.container.VerticalContainer(widget, width = 10, height = 400, top = 460, left = 920)
    self.arrow_box = pudding.control.Box(widget, width = 70, height = 240, top = 250, left = 900, \
        background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .3))
    self.box = pudding.control.Box(widget, width = 960, height = 280, top = 230, left = 30, \
      background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .1))
    self.text = pudding.control.SimpleLabel(widget, label = "select a world", font = font1, top = 200, left = 30, color = (1, 1, 1, .5))
    self.arrow_down = self.down_cont.add_child(text_menu('down', self.down, font1, None))
    self.arrow_up = self.up_cont.add_child(text_menu(' up', self.up, font1, None))
  def menu(self, date):
    try:
      for i in scene.recursive():
        if type(i) == soya.Light:
          scene.remove(i)
    except:
      print "no sun in world"
      pass
    try:
      preview.advance_time = lambda r: self.preview.rotate_y(0*r) # reset rotation to 0 speed 
      for i in preview.recursive():
        preview.remove(i)
      try: # remove plants if they were generated. must be removed separately.
        preview.remove(plant_world)
      except:
        print "no plants found"
        pass
    except:
      pass
    try:
      scene.remove(preview)
    except:
      pass
    # make a check for no existing worlds here or put in a demo world
    self.setup()
    visibility([back_label, back_cont, logo, self.cont, self.box, self.text, self.down_cont, self.up_cont, \
    self.arrow_box, self.arrow_down, self.arrow_up], 1)
    visibility([menu_box, menu_cont], 0)
    self.date = date
    width = 320
    height = 240
    offset = 0
    os.path.walk(os.path.join(PG_DIR, 'data', 'images', 'thumbnails'), self.get_pics, 1)
    pad = 250
    for i in self.pics[0]:
      thumbnail = pil.open(os.path.join(PG_DIR, 'data', 'images', 'thumbnails', i))
      thumbnail.anchors = pudding.ANCHOR_ALL
      self.cont.add_child(pudding.ext.slicingimage.SlicingImage(self.cont, thumbnail, top=offset +pad, left=50,  z_index=-3))
      date =  str(i).strip('pg_').strip('.jpeg')
      try: # create stats labels
        stats_file = open(os.path.join(PG_DIR, 'stats', 'pg_stats_' +date +'.sheet'), 'r')
        stats_txt = stats_file.read()
        self.cont.add_child(pudding.control.SimpleLabel(self.cont, label = str(stats_txt), \
          autosize = True, font=font1, top=offset+pad, left=400, color= (1, 1, 1, .3)))
      except:
        pass
      self.cont.add_child(pudding.control.SimpleLabel(self.cont, label = str(date),  \
        autosize = True, font=font1, top=offset+370, left=55, color= (1, 1, 1, .3), z_index=-3))
      self.cont.add_child(button_menu(320, 240, offset +pad, 50, gen.previewer, "", (date, 1))) # load chosen world 
      offset += (height+500)
  def up(self, date):
    if self.cont.top < 0:
      self.cont.top +=740
    else:
      self.cont.top = (len(self.pics[0])-1)*-740
  def down(self, date):
    if self.cont.top > (len(self.pics[0])-1)*-740:
      self.cont.top -=740
    else:
      self.cont.top = 0

class plant_guide:
  '''
  interface to the guide
  '''
  def __init__(self):
    atmosphere = soya.SkyAtmosphere()
    atmosphere.ambient = (0.3, 0.3, 0.3, 1)
    atmosphere.skyplane  = 1
    atmosphere.sky_color = (1, 1, 0.8, 1)
    scene.atmosphere = atmosphere
    self.setup()
  def get_pics(self, arg, top, names):
    self.pics.append(names)
  def setup(self):
    self.pics = []
    self.cont = pudding.container.VerticalContainer(widget, width = 800, height = 400) # this container is moved up and down
    self.up_cont = pudding.container.VerticalContainer(widget, width = 10, height = 400, top = 260, left = 920)
    self.down_cont = pudding.container.VerticalContainer(widget, width = 10, height = 400, top = 460, left = 920)
    self.arrow_box = pudding.control.Box(widget, width = 70, height = 240, top = 250, left = 900, \
        background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .3))
    self.box = pudding.control.Box(widget, width = 960, height = 280, top = 230, left = 30, \
      background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .1))
    self.text = pudding.control.SimpleLabel(widget, label = "packet garden protocol guide", font = font1, top = 200, left = 30, color = (1, 1, 1, .5))
    self.arrow_down = self.down_cont.add_child(text_menu('down', self.down, font1, None))
    self.arrow_up = self.up_cont.add_child(text_menu(' up', self.up, font1, None))
  def menu(self, date):
    self.setup()
    visibility([back_label, back_cont, logo, self.cont, self.box, self.text, self.down_cont, self.up_cont, \
    self.arrow_box, self.arrow_down, self.arrow_up], 1)
    visibility([menu_box, menu_cont], 0)
    self.date = date
    width = 320 
    height = 240
    offset = 0 
    os.path.walk(os.path.join(PG_DIR, 'data', 'images', 'guide'), self.get_pics, 1)
    pad = 240
    for i in self.pics[0]:
      img = pil.open(os.path.join(PG_DIR, 'data', 'images', 'guide', i))
      img.anchors = pudding.ANCHOR_ALL
      self.cont.add_child(pudding.ext.slicingimage.SlicingImage(self.cont, img, top=offset +pad, left=43,  z_index=-3))
      name =  str(i)[:-4]
      try: 
        guide_file = open(os.path.join(PG_DIR, 'guide', name +'.guide'), 'r')
        guide_txt = guide_file.read()
        self.cont.add_child(pudding.control.SimpleLabel(self.cont, label = str(guide_txt), \
          autosize = True, font=font1, top=offset+pad, left=310, color= (1, 1, 1, .3)))
      except:
        pass
      offset += (height+500)
  def up(self, date):
    if self.cont.top < 0:
      self.cont.top +=740
    else:
      self.cont.top = (len(self.pics[0])-1)*-740
  def down(self, date):
    if self.cont.top > (len(self.pics[0])-1)*-740:
      self.cont.top -=740
    else:
      self.cont.top = 0

class pg_label(soya.World):
  '''
  create in game ip labels.
  calls laser class (see below)
  '''
  def __init__(self, world, x, y, z, ip, camera, player, toggle):
    soya.World.__init__(self, world)
    self.camera = camera
    self.player = player
    self.toggle = toggle
    self.x, self.y, self.z = x, y, z
    self.world = world
    self.track = soya.Point(self.world, 0.0, 0.0, 0.0)
    self.dir = dir
    ip = ip.strip()
    ip = ip.strip("'")
    self.ip = ip
    '''
    if 'home' in self.ip:
      mat = soya.Material()
      mat.texture = soya.Image.get("glint.png")
      aura=soya.particle.Smoke(self.world)
      aura.material = mat
      #aura.set_colors((1.0, 1.0, 1.0, 1.0), (0.2, 0.7, 0.8, 0.8), (0.2, 0.4, 0.8, 0.2))
      aura.set_colors((0.1, 0.1, 0.1, 1.0), (0.3, 0.3, 0.3, 1.0), (0.3, 0.3, 0.3, 1.0), (0.1, 0.1, 0.1, 1.0))
      #aura.set_colors((0.1, 0.1, 0.1, 1.0), (0.3, 0.3, 0.3, 1.0), (0.3, 0.3, 0.3, 1.0), (0.1, 0.1, 0.1, 1.0))
      #p.set_colors((0.5,0.5,0.0,1), (0.5,0.5,0.0,1),(0.0,0.0,0.5,1),(0.5,0.0,0.0,1),(0.0,0.0,0.0,0.5))
      #aura.set_sizes ((0.19, 0.19), (0.35, 0.35))
      aura.nb_particles = 5
      aura.max_particles_per_round=50
      aura.set_xyz(self.x * 30, self.y * 30, self.z * 30)
      aura.add_mul_vector(-4.2, self.vector_to(self.world.position()))
      aura.auto_generate_particle=1
    else:
      pass
    '''
    self.label3d = soya.label3d.Label3D(self.world, self.ip)
    self.label3d.set_xyz(self.x *30, self.y*30, self.z *30)
    self.label3d.visible = 1
    self.label3d.look_at_y(self.world.position().vector_to(self.label3d.position()))
    # lasers look good but they really hit the CPU hard. 
    #if self.toggle == 1:
      #self.laser = pg_laser(center.position(), self.parent, 0.0, 0.0, 0.0)
      #self.laser.look_at(self.label3d.vector_to(center.position()))
    #else:
    #  pass
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    #pd = self.label3d.position().distance_to(self.player.position())
    cd = self.label3d.position().distance_to(self.camera.position())
    self.label3d.look_at(self.camera)
    if self.toggle == 1:
      if cd < 20 and cd > 4:
        self.label3d.visible = 1
      # self.laser.x = self.label3d.x
      # self.laser.y = self.label3d.y
      # self.laser.z = self.label3d.z
      # self.laser.visible = 1
      else:
        self.label3d.visible = 0
      # self.laser.visible = 0
    else:
      if cd < 18 and cd > 4: 
        self.label3d.visible = 1
      else:
        self.label3d.visible = 0

class pg_laser(soya.laser.Laser):
  '''
  laser class called by pg_label (above).
  '''
  def __init__(self, target, world, x, y, z):
    soya.laser.Laser.__init__(self, world)
    self.x, self.y, self.z = x, y, z 
    self.target = target.position()
    self.world = world
    self.vector = self.position().vector_to(target)
    self.look_at(self.vector)
    self.world.add(self)
    self.visible = 0
    self.color = (0.4, 0.4, 0.8, 0.15)# light blue grey

class make_plant(soya.World):
  ''' 
  simple plant builder class. this will become more sophisticated in future versions.
  '''
  def __init__(self, world, x, y, z, proto, scale, suffix):
    self.suffix = suffix
    soya.World.__init__(self, world)
    self.proto = proto
    self.world = world
    self.plant = soya.World()
    self.x, self.y, self.z = x, y, z
    self.select_plant(self.proto, scale, self.suffix)
  def select_plant(self, proto, scale, suffix):
    leaf_num = random.choice((4, 5, 6, 7))
    if proto in ['ssh', 'https']:
      self.build_plant("ssh", 3, leaf_num, scale)
    elif proto == 'ftp':
      self.build_plant("ftp", 3, leaf_num, scale)
    elif proto in ['smtp', 'pop3', 'imap']:
      self.build_plant("mail", 4, leaf_num, scale)
    elif proto in ['gizmo', 'icq', 'msn']: 
      self.build_plant("im", 4, leaf_num, scale)
    elif proto in ['soulseek', 'mule', 'bt', 'gnutella']:
      self.build_plant("p2p", 7, leaf_num, scale)
    elif proto in ['ut2004', 'quake3']:
      self.build_plant("game", 7, leaf_num, scale)
    elif proto == 'http':
      self.build_plant('None', 0, leaf_num, scale)
    elif proto == 'None':
      self.build_plant('home', 0, leaf_num, scale)
    else:
      self.bloom_mdl = None
    self.plant.set_xyz(self.x *30, self.y *30, self.z *30)
    vec = self.plant.position().vector_to(self.world.position()) 
    self.plant.look_at(vec)
    self.world.add(self.plant)
  def build_plant(self, mdl, bloom_count, leaf_count, scale):
    # setup blooms.
    scale = scale*15
    self.bloom_count = bloom_count
    if mdl == 'None':
      self.bloom_mdl = None
    elif mdl == 'home':
      self.bloom_mdl = soya.World.get(mdl +self.suffix)
      mat = soya.Material()
      mat.texture = soya.Image.get("home_leaf.png")
      mat.environment_mapping = 1
      for face in self.bloom_mdl.children:
        face.material=mat
        face.double_sided = 1
      self.bloom = soya.Body(self.plant, self.bloom_mdl.to_model())
      self.bloom.rotate_lateral(random.choice((1, 2, 3, 4))*90)
      self.bloom.solid = 0
      self.bloom.scale(scale, scale, scale)
      self.bloom.rotate_vertical(90)
    elif mdl == 'ssh':
      self.bloom_mdl = soya.World.get(mdl +self.suffix)
      mat = soya.Material()
      mat.texture = soya.Image.get("home_leaf.png")
      mat.environment_mapping = 1
      for face in self.bloom_mdl.children:
        face.material=mat
        face.double_sided = 1
      for i in range(bloom_count):
        self.bloom = soya.Body(self.plant, self.bloom_mdl.to_model())
        self.bloom.z -= 4/(random.choice((2, 3, 4, 5)))
        self.bloom.rotate_lateral(random.choice((1, 2, 3, 4))*90)
        self.bloom.solid = 0
        self.bloom.scale(scale, scale, scale)
        self.bloom.rotate_vertical(90)
    else:
      # don't create bloom body here, won't work.
      self.bloom_mdl = soya.Shape.get(mdl + self.suffix)
      if self.bloom_count > 1:
        for i in range(bloom_count):
          self.bloom = soya.Body(self.plant, self.bloom_mdl)
          self.bloom.z -= 4/(random.choice((2, 3, 4, 5)))
          self.bloom.rotate_lateral(random.choice((1, 2, 3, 4))*90)
          self.bloom.solid = 0
          self.bloom.scale(scale, scale, scale)
          self.bloom.rotate_vertical(90)
      else:
        self.bloom = soya.Body(self.plant, self.bloom_mdl)
        self.bloom.rotate_lateral(random.choice((1, 2, 3, 4))*90)
        self.bloom.solid = 0
        self.bloom.scale(scale, scale, scale)
        self.bloom.rotate_vertical(90)
    # setup leaves
    if mdl == 'home':
      mat = soya.Material()
      mat.texture = soya.Image.get("home_leaf.png")
      mat.environment_mapping = 1
      for i in range(10):
        #self.leaf = soya.Body(self.plant, self.leaf_mdl)
        self.leaf_world = soya.World.get('home_leaf'+self.suffix)
        self.leaf_world.material = mat
        age = random.choice((.6, .8, 1.0))*1
        for face in self.leaf_world.children:
          face.material=mat
          face.double_sided = 1
        self.leaf = soya.Body(self.plant, self.leaf_world.to_model())
        self.leaf.scale(age, age, age)
        self.leaf.solid = 0
        self.leaf.rotate_lateral(random.choice((1, 2, 3, 4, 5, 6, 7, 8))*45)
        self.leaf.z = (random.random()*(-4*scale/15))
        self.leaf.rotate_vertical(90)
        self.leaf.scale(scale, scale, scale)
    # setup stem
      self.stem_world = soya.World.get("home_stem"+self.suffix)
      for face in self.leaf_world.children:
        face.material=mat
      self.stem = soya.Body(self.plant, self.stem_world.to_model())
      self.stem.solid = 0
      self.stem.scale(scale, scale, scale)
      self.stem.rotate_vertical(90)
    else:
      self.leaf_mdl = soya.Shape.get("leaf" +self.suffix)
      for i in range(leaf_count):
        self.leaf = soya.Body(self.plant, self.leaf_mdl)
        self.leaf.solid = 0
        self.leaf.z = (random.random()*(-4*scale/15))
        self.leaf.scale(scale, scale, scale)
        age = random.choice((.6, .8, 1.0))*1
        self.leaf.rotate_lateral(random.choice((1, 2, 3, 4, 5, 6, 7, 8))*45)
        self.leaf.scale(age, age, age)
        #self.leaf.rotate_lateral(random.random()*360)
        self.leaf.rotate_vertical(90)
    # setup stem
      self.stem_mdl = soya.Shape.get("stem" +self.suffix)
      self.stem = soya.Body(self.plant, self.stem_mdl)
      self.stem.solid = 0
      self.stem.scale(scale, scale, scale)
      self.stem.rotate_vertical(90)
    #self.plant.to_model()
class capture_interface:
  '''
  interface to choose network device on UNIX systems
  Tested on Debian GNU/Linux only.
  '''
  def __init__(self):
    attempts = 0
    self.nic_list = []
    self.capture_txt = "start capture"
    self.device_txt = ""
    self.setup()
  def setup(self):
    self.cont = pudding.container.VerticalContainer(widget, left = 10, top = 210)
    self.cont.bottom = 10
    self.cont.right = 10
    self.send_label = pudding.control.SimpleLabel(self.cont, label = "start capture", font=font1, top=78, left=60)
    self.device_label = pudding.control.SimpleLabel(self.cont, label = self.device_txt, font=font1, top=45, left=35)
    if "win32" in PLATFORM:
      pass
    else:
      self.stop_label = pudding.control.SimpleLabel(self.cont, label = "stop capture", font=font1, top=98, left=60)
      self.stop_label.visible = 0
    self.device_label.color = (1, 1, 1, .4)
    self.box = pudding.control.Box(self.cont, width = 960, height = 200, top = 20, left = 20, \
        background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .1))
  def menu(self, arg):
    self.setup()
    time.sleep(1)
    menu_toggle(0)
    self.send_label.visible = 0
    try:
      self.cap_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      self.cap_sock.connect(('localhost', 2080))
      self.device_label.label = "it appears packet garden is already capturing traffic.\ntry generating a world."
      back_cont.visible = 1
      back_label.visible = 1
      if "win32" in PLATFORM:
        pass
      else:
        self.send_stop = self.cont.add_child(button_menu(128, 30, 92, 35, self.stop, "", None))
        self.stop_label.visible = 1
    except:
      user_cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'r')
      options = user_cfg.read()
      if re.search("fullscreen on", options):
        self.device_label.label ="packet capture cannot be started in full screen mode.\nplease use the configure menu to disable fullscreen mode and try starting packet capture again."
      else:
        try:
          if "win32" in PLATFORM:
            print self.detect_online()
            if self.detect_online() == 0:
              raise Exception
            else:
              self.send_label.visible = 1
              self.send = self.cont.add_child(button_menu(128, 30, 72, 35, self.start, "", None))
              self.send.visible = 1
          else:
            self.send_label.visible = 1
            self.device = self.detect_online()
            self.device_label.label = "it appears the device " +self.device +" is connected to the internet. going with this."
            self.send = self.cont.add_child(button_menu(128, 30, 72, 35, self.start, "", None))
            self.send.visible = 1
        except:
          self.device_label.label = "it appears you're not connected to the internet. try to get online and restart Packet Garden.\nsee the file README.txt for more information."  
          self.device_label.top = 30
      visibility([menu_cont, menu_box, selector.up_cont, selector.down_cont, selector.arrow_box, selector.cont, selector.box, selector.text], 0)
      visibility([capture.cont, self.box, back_label, back_cont], 1)
  def start(self, arg):
    if "win32" in PLATFORM:
      try:
        cap = subprocess.Popen('pg_packet.exe')
        menu_toggle(1)
      except:
        self.device_label.label = "tried to start but couldn't. either you're not online or packet garden is already capturing traffic."
        pass   
    elif "darwin" in PLATFORM:
      auth = Authorization(destroyflags=(kAuthorizationFlagDestroyRights,))
      cmd = ('/usr/local/bin/pythonw2.5',os.path.join(os.path.dirname(__file__),'pg_packet.py'),self.device,PG_DIR)
      cap = auth.executeWithPrivileges(*cmd)
      menu_toggle(1)
    else:
      try:
        subprocess.Popen('sudo -k', shell=True)
      except:
        pass
      time.sleep(1)
      self.send.visible = 0
      self.send_label.visible = 0
      try:
        cap = subprocess.Popen("gksudo --message='Packet Garden needs permission to read information from your network card.' python /usr/share/games/packetgarden/pg_packet.py " +self.device +" " +PG_DIR, shell=True)
      except:
        cap = subprocess.Popen("gksudo --message='Packet Garden needs permission to read information from your network card.' python ./pg_packet.py " +self.device +" " +PG_DIR, shell=True)
      menu_toggle(1)
      try:
        subprocess.Popen('sudo -k', shell=True)
      except:
        pass
  def stop(self, arg):
    # stopping packet capture is currently not working on Windows. will probably move to using sockets to send an exit msg to \
    # capture script as killing processes is too rough.
    if "win32" in PLATFORM:
      pass
    elif "darwin" in PLATFORM:
      auth = Authorization(destroyflags=(kAuthorizationFlagDestroyRights,))
      cmd = os.path.join(os.path.dirname(__file__),'stop_capture')
      cap = auth.executeWithPrivileges(cmd)
    else:
      try:
        subprocess.Popen('sudo -k', shell=True)
      except:
        pass
      try:
        subprocess.Popen("gksudo --message='Packet Garden needs permission to stop the capture process.' /usr/share/games/packetgarden/stop_capture", shell=True)
      except:
        subprocess.Popen("gksudo --message='Packet Garden needs permission to stop the capture process.' ./stop_capture", shell=True)
    menu_toggle(1)
  def detect_online(self):
    # thanks to pix for suggesting route as an easier solution than regexing ifconfig and trying a ping.
    if "darwin" in PLATFORM:
      route_tab = os.popen('netstat -rn').readlines()
      for lines in route_tab:
        if re.search('default', lines):
          device = lines.split()[5:][0]
          return device
    elif "win32" in PLATFORM:
      strComputer = "."
      objWMIService = win32com.client.Dispatch("WbemScripting.SWbemLocator")
      objSWbemServices = objWMIService.ConnectServer(strComputer,"root\cimv2")
      colItems = objSWbemServices.ExecQuery("Select * from Win32_NetworkAdapterConfiguration")
      a = 0
      for objItem in colItems:
        z = objItem.DefaultIPGateway
        if z is None:
          pass
        else:
          for x in z:
            ip_gateway = x
            a = 1
            break
      return a
    else:
      route_tab = os.popen('/sbin/route -n').readlines()
      for lines in route_tab:
        if re.search('UG', lines) and re.search('0.0.0.0', lines):
          device = lines.split()[7:][0]
          return device
#function that will attempt to find external IP, defers to $HOSTNAME
def get_external_ip():
  try: 
    hostip = url.open("http://www.hostip.info/index.html", 'r')
    for lines in hostip:
      if re.findall("Your IP Address", lines):
        a, b = lines.split(':')
        ext_ip, c = b.split('<')
    return ext_ip, 1
  except:
    return hostname, 0

def visibility(list, toggle):
  for obj in list:
    obj.visible = toggle

def go_quit(date):
  print "quitting..."
  raise SystemExit()

def menu_toggle(arg):
  if arg == 1:
    visibility([gen.cont], 0)
    visibility([menu_cont, menu_box], 1)
    visibility([configure.cont, selector.up_cont, selector.down_cont, selector.arrow_box, \
      selector.cont, selector.box, selector.text, back_label, back_cont, configure.cont, configure.slider_label, \
      configure.slider_puck, capture.cont, guide.cont, guide.text, guide.box, guide.arrow_box, guide.down_cont, \
      guide.up_cont, gen.cont], 0)
  elif arg == 0:
    visibility([menu_cont, menu_box], 0)
  #pudding.main_loop.MainLoop(scene).idle()

# create our root widget
widget = pudding.core.RootWidget(width = 1024,height = 768)

# setup fonts 
font1 = soya.Font(os.path.join(PG_DIR, 'data', 'FreeSans.ttf'), 14, 10)
font2 = soya.Font(os.path.join(PG_DIR, 'data', 'FreeSans.ttf'), 20, 10)
font3 = soya.Font(os.path.join(PG_DIR, 'data', 'FreeSans.ttf'), 16, 16)

# create containers
menu_cont = pudding.container.VerticalContainer(widget, width = 10, height = 50, left = 50, top = 240, z_index=-4)
back_cont = pudding.container.VerticalContainer(widget, width = 10, height = 50, left = 30, top = 650, z_index=-4)
back_label = pudding.control.SimpleLabel(widget, label = "main menu", font=font1, top=655, left=50)
play_cont = pudding.container.VerticalContainer(widget, width = 10, height = 50, left = 180, top = 700, z_index=-4)
play_label = pudding.control.SimpleLabel(widget, label = "visit world", font=font1, top=705, left=200)
quit_cont = pudding.container.VerticalContainer(widget, width = 10, height = 50, left = 30, top = 700, z_index=-4)
quit_label = pudding.control.SimpleLabel(widget, label = "quit", font=font1, top=705, left=50)
menu_box = pudding.control.Box(widget, width = 960, height = 200, top = 230, left = 30, \
  background_color = (0, 0, 0, .1), border_color = (1, 1, 1, .1))
game_box = pudding.control.Box(widget, width = 1004, height = 80, top = 678, left = 10, \
        background_color = (0, 0, 0, .2), border_color = (1, 1, 1, .1))
# game msg
game_msg = \
  "move around using the cursor keys" '\n' \
  +"take a screenshot using the S key" '\n' \
  +"view world in wireframe mode with the W key" '\n' \
  + "hide IP labels with the L key and plants with the P key" '\n' \
  +"ESC quits"
game_label = pudding.control.SimpleLabel(widget, label = game_msg, \
  autosize = True, font=font1, top=680, left=30, color= (1, 1, 1, .2))

# logo
logo_pic = pil.open(os.path.join(PG_DIR, 'data', 'images', 'pg_logo.png')) 
logo = pudding.ext.slicingimage.SlicingImage(widget, logo_pic, top=645, left=600) # we need to reach this later

# create core interface instances
gen = make_world()
configure = configurator()
capture = capture_interface()
selector = world_selector()
guide = plant_guide()

# create main menu entries
menu_cont.add_child(text_menu('grow a garden from todays traffic', gen.generate, font1, None))
menu_cont.add_child(text_menu('start and stop packet capture', capture.menu, font1, None))
menu_cont.add_child(text_menu('visit past gardens', selector.menu, font1, date))
menu_cont.add_child(text_menu('protocol guide', guide.menu, font1, None))
menu_cont.add_child(text_menu('configure', configure.menu, font1, None))
back_cont.add_child(button_menu(128, 30, 0, 0, menu_toggle, "", 1))
play = play_cont.add_child(button_menu(128, 30, 0, 0, gen.go_world, "", None)) # we need to reach this later
quit_cont.add_child(button_menu(128, 30, 0, 0, go_quit, "", 1))

# turn off default stuff
menu_toggle(1)

visibility([game_label, selector.cont, selector.box, selector.text, selector.down_cont, selector.up_cont, \
selector.arrow_box, selector.arrow_down, selector.arrow_up, game_box, play_cont, play_label, back_cont, \
back_label, capture.cont, guide.cont, guide.text, guide.box, guide.arrow_box, guide.down_cont, guide.up_cont], 0)

# add our root widget
widget.add_child(camera)
soya.set_root_widget(widget)
pudding.main_loop.MainLoop(scene).idle()
