#! /usr/bin/python -O

# Packet Garden capture script (beta). Cross platform version. 
# 
# Copyright (C) 2006 Julian Oliver 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import pcap
import dpkt
import time
import socket
import os 
import os.path
import sys
import gzip

try:
  import psyco
  psyco.full()
except ImportError:
  pass
try:
  import struct
except ImportError:
  pass
try:
  import fcntl
except ImportError:
  pass
try:
  import win32com.client
except ImportError:
  pass
try:
  import GeoIP
except ImportError:
  print "GeoIP module not found. there will be no country codes generated for captured IP's"
  pass

# setup wide scope stuff 
dict = {}
hostname = socket.gethostname()
msg = "-----------------------------------------------" '\n' \
      "| Invalid device or insufficient permissions. |" '\n' \
      "| Become root or try another device.          |" '\n' \
      "-----------------------------------------------" 

PLATFORM = sys.platform

# a dictionary of ports to filter for. add to these lists and/or make new ones.
# i've only included ports used by software available on Linux, OS X and Windows

ports = {
        'bt' :        [i for i in range(10000, 10010)] + [6881],
        'mule':       [4662, 4672],
        'gnutella' :  [6346],
        'soulseek' :  [2240],
        'gizmo' :     [5004, 5005, 64064],
        'icq' :       [i for i in range(20000, 20019)] + [5190],
        'yahoo' :     [i for i in range(5000, 5001)]+ [5050],
        'msn' :       [1863, 6901],
        'icecast' :   [i for i in range(8000, 8001)],
        'quake3' :    [27950, 27952, 27960, 27965],
        'ut2004' :    [i for i in range(7777, 7788)] + [27900, 42292],
        'imap' :      [143],
        'pop3' :      [119, 110, 995],
        'ssh' :       [22],
        'ftp' :       [20, 21],
        'https' :     [443],
        'http' :      [80]
        }

if "win32" in PLATFORM:
  strComputer = "."
  objWMIService = win32com.client.Dispatch("WbemScripting.SWbemLocator")
  objSWbemServices = objWMIService.ConnectServer(strComputer,"root\cimv2")
  colItems = objSWbemServices.ExecQuery("Select * from Win32_NetworkAdapterConfiguration")
# packet capture class
class packet:
  def get_local_ip(self, device): 
    if "win32" in PLATFORM: # shennanegans for Windows.
      a = 0
      for objItem in colItems:
          z = objItem.DefaultIPGateway
          if z is None:
              a = 1
          else:
              for x in z:
                  ip_gateway = x
          z = objItem.IPAddress
          if z is None:
              a = 1
          else:
              for x in z:
                  ip_addr = x
                  return ip_addr
    elif "darwin" in PLATFORM: # OS X
      ifconfig = os.popen('ifconfig ' + device).readlines()
      for lines in ifconfig:
        if "inet" and "netmask" in lines:
          ip_addr = lines.split()[1]
          return ip_addr
    else: # defaults to Linux
      s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
      return socket.inet_ntoa(fcntl.ioctl(s.fileno(), 0x8915, struct.pack('256s', device[:15]))[20:24])
  def geo_ip(self, ip):
    if "win32" in PLATFORM:
      return ""
    else:
      gi = GeoIP.new(GeoIP.GEOIP_MEMORY_CACHE)
      return gi.country_code_by_addr(str(ip[0])+"."+str(ip[1])+"."+str(ip[2])+"."+str(ip[3]))
  def capture(self, device, dir):
    self.dir = dir
    #we're running this script as root therefore we don't want to expand the user from "~/".
    PG_DIR = self.dir
    start = time.strftime('%d'+'-'+'%m'+'-'+'%y')
    # read in old log if it exists. 
    # can't just append to it as we need to collate with existing entries.
    try:
      old_log = gzip.open(os.path.join(PG_DIR, 'logs', 'pcap' +'_' +hostname +'_'+start +'.log.gz'), 'r')
      for lines in old_log:
        ip, dir, proto, region, length = lines.split(',')
        ip = ip.lstrip("('")
        dir = dir.strip()
        region = region.strip().rstrip("'")
        length = int(length.strip().rstrip(')'))
        dict[str(ip) +", " +dir +", " +proto +", " +region] = length
    except:
      pass
    self.device = device
    count = 0 
    if "win32" in PLATFORM:
      p = pcap.pcap()  
      ip = self.get_local_ip(None)
    else:
      p = pcap.pcap(self.device)
      ip = self.get_local_ip(self.device)
    list = []
    ip = b1, b2, b3, b4 = tuple([int(n) for n in ip.split('.')])
    net = (b1, b2, b3)
    while 1:
      try:
        for ts, pkt in p:
          packet = dpkt.ethernet.Ethernet(pkt)
          source = struct.unpack('4B', packet.data.src)
          dest = struct.unpack('4B', packet.data.dst) # unpack 4 block struct type.
          length = packet.data.len
          source_port = packet.data.data.sport
          dest_port = packet.data.data.dport
          geo_source = self.geo_ip(source)
          geo_dest = self.geo_ip(dest)
          # we're only interested in UDP and TCP packets
          if type(packet.data.data) == dpkt.tcp.TCP or type(packet.data.data) == dpkt.udp.UDP:
            count += 1
            # filter and tag on direction
            if source == ip: 
              dir = 'up'
            elif dest == ip:
              dir = 'down'
            # filter and tag on port by iterating through ports dict. 
            for k, v in ports.iteritems():
              if source_port in v or dest_port in v:
                proto = k
                break
              else:
                proto = 'other'
            # don't log packets from sources on the local network
            if (source[0], source[1], source[2]) == (b1, b2, b3) and source != ip:
              pass
            elif (source[0], source[1], source[2]) == (b1, b2, b3) and dest != ip:
              pass
            if source == ip:
              dest_net = str(dest[0]) +"." +str(dest[1]) +"." +str(dest[2]) +"." +str(dest[3])
              # fast collation using key matching. avoids expensive iterations.
              try:
                dict[str(dest_net) +", "  +dir +", " +proto +", " +geo_dest] += length
              except:
                dict[str(dest_net) +", "  +dir +", " +proto +", " +geo_dest] = length
            else:
              source_net = str(source[0]) +"." +str(source[1]) +"." +str(source[2]) +"." +str(source[3])
              try:
                dict[str(source_net) +", " +dir +", " +proto +", " +geo_source] += length
              except:
                dict[str(source_net) +", " +dir +", " +proto +", " +geo_source] = length
            if count > 1000:
              count = 0
              current = time.strftime('%d'+'-'+'%m'+'-'+'%y')
              t = time.time() 
              if current == start:
                log = gzip.open(os.path.join(PG_DIR, 'logs', 'pcap' +'_' +hostname +'_'+start +'.log.gz'), 'w')
                for i in dict.items():
                  log.write(str(i) +'\n')
              else: # accomodate for day changeover. there'll always be a few lost packets.
                dict.clear()
                start = current
                log = gzip.open(os.path.join(PG_DIR, 'logs', 'pcap' +'_' +hostname +'_'+start +'.log.gz'), 'w')
              log.flush()
              log.close()
              total = time.time() - t
              print "processed ", len(dict), "networks in ", total, "seconds" 
            else:
              pass
          else:
            print "dead packet"
      except:          
        pass

cap = packet()
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
sock.bind(('localhost', 2080)) # this is only visible to clients on the same machine
sock.listen(5)

if "win32" in PLATFORM:
  device = None
  dir = "" 
  cap.capture(device, dir)
elif "darwin" in PLATFORM:
  #try:
    device = sys.argv[1]
    device.strip()
    cap.capture(device, dir)
  #except:
    print msg
    raise SystemExit()
else:
  try:
    device = sys.argv[1]
    dir = sys.argv[2]
    device.strip(); dir.strip()
    cap.capture(device, dir)
  except:
    print msg
    raise SystemExit()

