#! /usr/bin/python -O

# Packet Garden player script (beta). Cross platform version. 
# 
# Copyright (C) 2006 Julian Oliver
#
# This script is really a modification of a Soya tutorial file, adadpted
# to simulate rolling around a spheroid terrain. I added a torch, a ray effect, 
# quit and restart and a screenshot facility.
# 
# License:
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, os, os.path, math
import soya
import soya.label3d
import soya.sdlconst as sdlconst
import soya.ray
import socket
import time
import re

try:
  import psyco
  psyco.full()
except ImportError:
  pass

if "darwin" in sys.platform or "win32" in sys.platform:
  PG_DIR = os.curdir 
else:
  PG_DIR = os.path.join(os.path.expanduser('~/'), '.packetgarden/')
soya.path.append(os.path.join(os.path.dirname(PG_DIR), "data"))

hostname = socket.gethostname()

class action:
  def __init__(self, action):
    self.action = action

ACTION_WAIT          = 0
ACTION_ADVANCE       = 1
ACTION_ADVANCE_LEFT  = 2
ACTION_ADVANCE_RIGHT = 3
ACTION_TURN_LEFT     = 4
ACTION_TURN_RIGHT    = 5
ACTION_GO_BACK       = 6
ACTION_GO_BACK_LEFT  = 7
ACTION_GO_BACK_RIGHT = 8

class kb_input:
  def __init__(self, world_shape, rotscene, label_world, plant_world, scene):
    self.world_shape = world_shape
    self.scene = scene
    self.plant_world = plant_world
    self.label_world = label_world
    self.rotscene = rotscene
    self.left_key_down = self.right_key_down = self.up_key_down = self.down_key_down = 0
    self.t_count = 0
    self.ip_count = 0
    self.plant_count = 0
    self.s_count = 0
    self.event = ()
  def next(self):
    for self.event in soya.process_event():
      if self.event[0] == sdlconst.KEYDOWN:
        if (self.event[1] == sdlconst.K_q) or (self.event[1] == sdlconst.K_ESCAPE):
          if "darwin" in sys.platform: 
            os.execl(os.path.join(os.path.dirname(__file__),'packetgarden'))
          
          elif "win32" in sys.platform:
            os.execl('pg_garden.exe')
          else:
            try:
              os.execl('/usr/share/games/packetgarden/pg_garden.py')
            except:
              os.execl('python ./pg_garden.py')
        elif self.event[1] == sdlconst.K_LEFT:  self.left_key_down  = 1
        elif self.event[1] == sdlconst.K_RIGHT: self.right_key_down = 1
        elif self.event[1] == sdlconst.K_UP:    self.up_key_down    = 1
        elif self.event[1] == sdlconst.K_DOWN:  self.down_key_down  = 1
        elif self.event[1] == sdlconst.K_w:     soya.toggle_wireframe()
        elif self.event[1] == sdlconst.K_s:  \
        soya.screenshot().resize((1024, 768)).save(os.path.join(PG_DIR, 'data', 'images', 'screenshots', 'pg_' +hostname +'_' +str(time.strftime('%S' +'%H'+'%M'+'_'+'%d'+'-'+'%h'+'-'+'%Y'))+'.jpeg'))
        if self.event[1] == sdlconst.K_l and self.ip_count == 0:  self.rotscene.remove(self.label_world); self.ip_count += 1
        elif self.event[1] == sdlconst.K_l and self.ip_count == 1: self.rotscene.add(self.label_world); self.ip_count = 0        
        if self.event[1] == sdlconst.K_p and self.plant_count == 0:  self.rotscene.remove(self.plant_world); self.plant_count += 1
        elif self.event[1] == sdlconst.K_p and self.plant_count == 1: self.rotscene.add(self.plant_world); self.plant_count = 0        
      elif self.event[0] == sdlconst.KEYUP:
        if   self.event[1] == sdlconst.K_LEFT:  self.left_key_down  = 0
        elif self.event[1] == sdlconst.K_RIGHT: self.right_key_down = 0
        elif self.event[1] == sdlconst.K_UP:    self.up_key_down    = 0
        elif self.event[1] == sdlconst.K_DOWN:  self.down_key_down  = 0

    return action({
      (0, 0, 1, 0) : ACTION_ADVANCE,
      (1, 0, 1, 0) : ACTION_ADVANCE_LEFT,
      (0, 1, 1, 0) : ACTION_ADVANCE_RIGHT,
      (1, 0, 0, 0) : ACTION_TURN_LEFT,
      (0, 1, 0, 0) : ACTION_TURN_RIGHT,
      (0, 0, 0, 1) : ACTION_GO_BACK,
      (1, 0, 0, 1) : ACTION_GO_BACK_LEFT,
      (0, 1, 0, 1) : ACTION_GO_BACK_RIGHT,
      }.get((self.left_key_down, self.right_key_down, self.up_key_down, self.down_key_down), ACTION_WAIT))

class avatar(soya.World):
  def __init__(self, parent, rotscene, world_shape, label_world, plant_world):
    soya.World.__init__(self, parent)
    self.label_world = label_world
    self.plant_world = plant_world
    self.world_shape = world_shape
    self.scene = parent
    self.rotscene = rotscene
    self.sphere = soya.World()
    input = kb_input(self.world_shape, rotscene, self.label_world, self.plant_world, self.scene)
    self.controller = input 
    self.ray_world = soya.World()
    self.speed = soya.Vector(self)
    self.y = 50 # don't touch
    self.solid = 0
    self.rotation_speed = 0.0
    self.sphere.center         = soya.Point(self.sphere, 0.0, 0.0, 0.0)
    self.radius         = 0.5
    self.radius_y       = 3.0
    self.center         = soya.Point(self, 0.0, self.radius_y, 0.0)
    self.left   = soya.Vector(self, -1.0,  0.0,  0.0)
    self.right  = soya.Vector(self,  1.0,  0.0,  0.0)
    self.up     = soya.Vector(self,  0.0,  1.0,  0.0)
    self.front  = soya.Vector(self,  0.0,  0.0, -1.0)
    self.back   = soya.Vector(self,  0.0,  0.0,  1.0)

    self.mdl = soya.Shape.get("player")
    self.body = soya.Volume(self, self.mdl)
    self.body.set_xyz(0.0, 0.5, 0.0)    
    self.body.scale(5, 5, 5)
    #self.body.material = material

    self.user_cfg = open(os.path.join(PG_DIR, 'config', 'pg_' +str(hostname) +'.config'), 'r')
    self.options = self.user_cfg.read()

    if re.search("shadows on", self.options):
      self.shadow_state = 1
      self.shadow_mdl = soya.Shape.get("player_shadow")
      self.body_shadow = soya.Volume(self, self.shadow_mdl)
      self.body_shadow.set_xyz(0.0, 0.5, 0.0)    
      self.body_shadow.scale(5, 5, 5)
      self.body_shadow.visible = 0 # an invisible volume still casts a shadow, which is what we want
    else:
      self.shadow_state = 0
      pass
    '''
    self.torch = soya.Light(self.parent)
    self.torch.constant = 2
    self.torch.angle = 90
    self.torch.diffuse = (.8, .8, 1, 1)
    '''
    self.shadow_color = (0.0, 0.0, 0.1, .4)
    self.ray = soya.ray.Ray(self.ray_world, length = 20)
    self.ray.z = -.5
    self.ray.x = -.5
    self.ray.endpoint = soya.Point(self.ray_world, 0.0, -1.95, 0.0)
    material = soya.Material()
    self.ray.material = material

    self.ray.visible = 1
    self.body.visible = 1
    
    material.diffuse = (0.0,0.0,1.0,1.0)
    material.additive_blending = 1 # Additive blending istypically used for FX
    self.parent.add(self.ray_world)

  def begin_round(self):
    self.begin_action(self.controller.next())
    soya.World.begin_round(self)
    
  def begin_action(self, action):
    self.rotation_speed = self.speed.x = self.speed.z = 0.0
    if self.speed.y > 0.0: self.speed.y = 0.0
    if  action.action in (ACTION_TURN_LEFT, ACTION_ADVANCE_LEFT):
      self.rotscene.rotate_lateral(-2)
    elif action.action in (ACTION_TURN_RIGHT, ACTION_ADVANCE_RIGHT):
      self.rotscene.rotate_lateral(2)
    if  action.action in (ACTION_ADVANCE, ACTION_ADVANCE_LEFT, ACTION_ADVANCE_RIGHT):
      self.rotscene.rotate_vertical(.5)
      self.body.rotate_vertical(-25)
    elif action.action == (ACTION_GO_BACK_LEFT):
      self.rotscene.rotate_lateral(-2)
      self.rotscene.rotate_vertical(-.5)
    elif action.action == (ACTION_GO_BACK_RIGHT):
      self.rotscene.rotate_lateral(2)
      self.rotscene.rotate_vertical(-.5)
    if action.action == (ACTION_GO_BACK):
      self.rotscene.rotate_vertical(-.5)
      self.body.rotate_vertical(25)
    new_center = self.center + self.speed

    self.down = self.center.vector_to(self.sphere.center)
    context = self.scene.RaypickContext(new_center, max(self.radius, 0.1 + self.radius_y))
    r = context.raypick(new_center, self.down, 0.1 + self.radius_y, 1, 1)
        
    if r:
      ground, ground_normal = r
      ground.convert_to(self)
      self.speed.y = ground.y

    else:
      self.speed.y = max(self.speed.y - 0.05, -0.5)
    for vec in (self.left, self.right, self.front, self.back, self.up):
      r = context.raypick(new_center, vec, self.radius, 1, 1)
      if r:
        collision, wall_normal = r
        hypo = vec.length() * self.radius - new_center.distance_to(collision)
        correction = wall_normal * hypo
        self.speed += correction
        new_center += correction
        
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    self.add_mul_vector(proportion, self.speed)
    self.ray_world.rotate_lateral(proportion *5)
    self.ray_world.set_xyz(self.x, self.y+1, self.z)
    '''
    self.torch.y = self.y+2
    self.torch.z = self.y-2
    '''
    if self.shadow_state == 1:
      self.body_shadow.set_xyz(self.body.x, self.body.y, self.body.z)
    else:
      pass
